import json
import os
from typing import Any, Dict, Union

from eth_keyfile import decode_keyfile_json
from eth_keys.datatypes import PrivateKey, PublicKey, Signature
from eth_utils import (
    keccak,
    to_bytes
)

from . import util


__all__ = [
    'keccak', 'recover_private_key', 'private_key_to_address',
    'create_key_pair', 'sign_hash', 'sig_to_address',
]


def recover_private_key(
        keystore: str, password: str, dump: bool = True
):
    """recover private key with keystore and password

    :param keystore: str, keystore -> json.dumps(keystore)
    :param password: str
    :param dump: bool, True: return `bytes.hex`, False: return bytes

    :return: str (bytes.hex) or bytes
    """
    keystore = json.loads(keystore)
    password = to_bytes(text=password)
    private_key = decode_keyfile_json(keystore, password)
    return util.dump_buf(private_key, dump)


def get_private_key(obj: Union[str, bytes, PrivateKey]) -> PrivateKey:
    """get private key from obj

    :param obj: str or bytes or eth_keys.datatypes.PrivateKey,
        raise exception if can not get private key

    :return: eth_keys.datatypes.PrivateKey
    """
    if isinstance(obj, PrivateKey):
        return obj

    if isinstance(obj, str):
        obj = to_bytes(hexstr=obj)

    if isinstance(obj, bytes):
        private_key = PrivateKey(obj)
        return private_key

    raise ValueError('can not get private key from %r' % obj)


def private_key_to_address(private_key: str) -> str:
    """return the address of public key

    :param private_key: str (hex str) or bytes or eth_keys.datatypes.PrivateKey

    :return: str, hex str, remove '0x' prefix
    """
    pk = get_private_key(private_key)
    address = pk.public_key.to_address()
    return util.remove_prefix_0x(address)


def keccak256(message: str) -> str:
    """message digest with keccak

    :param message: str
    :return: hex str
    """
    _hash = keccak(text=message)
    return _hash.hex()


def create_key_pair(dump: bool = True) -> Dict[str, Union[str, bytes]]:
    """create key pair

    :param dump: bool, True: return `bytes.hex`, False: return bytes

    :return: dict, {
            'privateKey': private_key, 'publicKey': public_key,
            'address': address
        }
    """
    # generate private key
    private_key = PrivateKey(os.urandom(32))
    public_key = PublicKey.from_private(private_key)
    address = private_key_to_address(private_key)

    return {
        'privateKey': util.dump_buf(private_key.to_bytes(), dump),
        'publicKey': util.dump_buf(public_key.to_bytes(), dump),
        'address': address,
    }


def sign_hash(_hash: str, private_key: str) -> Dict[str, str]:
    """sign hash with private_key

    :param _hash: message hash (digest message with keccak256)
    :param private_key: hex str, private key's hex str

    :return: dict, {'hash': _hash, 'signature': sign_hex}
    """
    # get signature
    pk = get_private_key(private_key)
    sign = pk.sign_msg_hash(util.hex_to_buf(_hash))
    sign_hex = util.remove_prefix_0x(sign.to_hex())
    return {
        'hash': _hash,
        'signature': sign_hex,
    }


def sig_to_address(msg_hash: str, sig: str) -> str:
    """get public key's address with msg_hash and sign hex str

    :param msg_hash: hex str of message hash
    :param sig: hex str of signature

    :return: hex str, public key's address
    """
    signature = Signature(
        signature_bytes=util.hex_to_buf(sig)
    )
    public_key = signature.recover_public_key_from_msg_hash(
        util.hex_to_buf(msg_hash)
    )
    address = public_key.to_address()
    return util.remove_prefix_0x(address)


def hash_text(message: str) -> str:
    """get the hash of text data

    :param message: str

    :return: hex str, digest message (keccak256)
    """
    return keccak256(message)


def hash_block_data(data: Dict[str, Any]) -> str:
    """get the hash of block data

    :param data: dict

    :return: hex str, digest message (keccak256)
    """
    sorted_data = util.get_sorted_qs(data)
    return keccak256(sorted_data)


def sign_block_data(data: Dict[str, Any], private_key: str) -> Dict[str, str]:
    """sign block data

    :param data: dict
    :param private_key: hex str (private key)

    :return: dict, {'hash': _hash, 'signature': sign_hex}
    """
    return sign_hash(hash_block_data(data), private_key)


def sign_text(message: str, private_key: str) -> Dict[str, str]:
    """sign text data

    :param message: str
    :param private_key: hex str, private key

    :return: dict, {'hash': _hash, 'signature': sign_hex}
    """
    return sign_hash(hash_text(message), private_key)


def sig_to_address_from_block(data: Dict[str, Any], sig: str) -> str:
    """get public key's address with block data and sign hex str

    :param data: dict, block data
    :param sig: hex str of signature

    :return: hex str, public key's address
    """
    return sig_to_address(hash_block_data(data), sig)



from collections import OrderedDict
from urllib.parse import urlencode, quote
from typing import Any, Dict, Union


def buf_to_hex(buf: bytes) -> str:
    return buf.hex()


def hex_to_buf(s_hex: str) -> bytes:
    return bytes.fromhex(s_hex)


def dump_buf(buf: bytes, dump: bool = True) -> Union[bytes, str]:
    return buf_to_hex(buf) if dump else buf


def quote_qs(
        string: str,
        safe: str = '',
        encoding: str = 'utf-8',
        errors: str = 'strict'
) -> str:
    return quote(string, safe=safe, encoding=encoding, errors=errors)


def get_sorted_qs(data: Dict[str, Any]) -> str:
    sorted_dict = OrderedDict(sorted(data.items()))
    # replace True => true, False => false
    for k, v in sorted_dict.items():
        if v is True:
            sorted_dict[k] = 'true'
        elif v is False:
            sorted_dict[k] = 'false'
    return urlencode(sorted_dict, quote_via=quote_qs)


def remove_prefix_0x(s: str) -> str:
    """remove prefix '0x' or '0X' from string s

    :param s: str

    :return: str, the substring which remove the prefix '0x' or '0X'
    """
    if s[:2].lower() == '0x':
        s = s[2:]
    return s



"""
    prs_utility

    utility for press.one Dapp developer.

    :copyright: © 2019 by the press.one team.
    :license: MIT, see LICENSE for more details.
"""
from .util import (
    remove_prefix_0x,
    buf_to_hex,
    hex_to_buf,
    dump_buf,
    quote_qs,
    get_sorted_qs,
)

from .core import (
    keccak256,
    get_private_key,
    recover_private_key,
    private_key_to_address,
    create_key_pair,
    sign_hash,
    sign_text,
    sign_block_data,
    sig_to_address,
    hash_text,
    hash_block_data,
    sig_to_address_from_block,
)

__version__ = '0.0.4'
__all__ = [
    'buf_to_hex',
    'hex_to_buf',
    'get_sorted_qs',
    'keccak256',
    'recover_private_key',
    'private_key_to_address',
    'create_key_pair',
    'hash_text',
    'hash_block_data',
    'sign_hash',
    'sign_text',
    'sign_block_data',
    'sig_to_address',
    'sig_to_address_from_block',
]



"""根据 keystore 和 password 得到私钥"""
import json

import prs_utility


keystore = {
    "address": "758ea2601697fbd3ba6eb6774ed70b6c4cdb0ef9",
    "crypto": {
        "cipher": "aes-128-ctr",
        "ciphertext": "92af6f6710eba271eae5ac7fec72c70d9f49215e7880a0c45d4c53e56bd7ea59",
        "cipherparams": {
            "iv": "13ddf95d970e924c97e4dcd29ba96520"
        },
        "mac": "b9d81d78f067334ee922fb2863e32c14cbc46e479eeb0acc11fb31e39256004e",
        "kdf": "pbkdf2",
        "kdfparams": {
            "c": 262144,
            "dklen": 32,
            "prf": "hmac-sha256",
            "salt": "79f90bb603491573e40a79fe356b88d0c7869852e43c2bbaabed44578a82bbfa"
        }
    },
    "id": "93028e51-a2a4-4514-bc1a-94b089445f35",
    "version": 3
}
password = '123123'


private_key = prs_utility.recover_private_key(
    json.dumps(keystore), password
)
print('private_key:', private_key)



"""根据 PRS 协议组合 block data, 并且使用 privateKey 进行签名"""
import prs_utility

with open(__file__) as fp:
    content = fp.read()

data = {
    'file_hash': prs_utility.keccak256(content),
}
key_pair = prs_utility.create_key_pair()
private_key = key_pair['privateKey']
sig = prs_utility.sign_block_data(data, private_key)
print('signature:', sig)



import json
import pytest

from prs_utility.util import (
    remove_prefix_0x, dump_buf, quote_qs, get_sorted_qs
)
from prs_utility.core import get_private_key
from prs_utility import (
    keccak256,
    recover_private_key,
    private_key_to_address,
    create_key_pair,
    sign_hash,
    sign_text,
    sign_block_data,
    sig_to_address,
    sig_to_address_from_block,
)
from prs_utility import hex_to_buf


KEYSTORE = {
    "address": "758ea2601697fbd3ba6eb6774ed70b6c4cdb0ef9",
    "crypto": {
        "cipher": "aes-128-ctr",
        "ciphertext": "92af6f6710eba271eae5ac7fec72c70d9f49215e7880a0c45d4c53e56bd7ea59",
        "cipherparams": {
            "iv": "13ddf95d970e924c97e4dcd29ba96520"
        },
        "mac": "b9d81d78f067334ee922fb2863e32c14cbc46e479eeb0acc11fb31e39256004e",
        "kdf": "pbkdf2",
        "kdfparams": {
            "c": 262144,
            "dklen": 32,
            "prf": "hmac-sha256",
            "salt": "79f90bb603491573e40a79fe356b88d0c7869852e43c2bbaabed44578a82bbfa"
        }
    },
    "id": "93028e51-a2a4-4514-bc1a-94b089445f35",
    "version": 3
}
PASSWORD = '123123'
PRIVATE_KEY = '6e204c62726a19fe3f43c4ca9739b7ffa37e4a3226f824f3e24e00a5890addc6'
ADDRESS = '758ea2601697fbd3ba6eb6774ed70b6c4cdb0ef9'


@pytest.mark.parametrize('s', ['0x123456', '0X123456', '123456'])
def test_remove_prefix_0x(s):
    remove_prefix_0x(s) == '123456'


@pytest.mark.parametrize(
    's, expected',
    [
        ('/', '%2F'),
        (' ', '%20'),
        ('+', '%2B'),
        ('http://a.com', 'http%3A%2F%2Fa.com'),
    ]
)
def test_quote_qs(s, expected):
    assert quote_qs(s) == expected


@pytest.mark.parametrize(
    'data, expected',
    [
        ({'a': 'A', 'b': True, 'c': False}, 'a=A&b=true&c=false'),
        ({'c': 'A', 'b': True, 'a': False}, 'a=false&b=true&c=A'),
        ({'c': 'a', 'b': 'b', 'a': 'c'}, 'a=c&b=b&c=a'),
    ]
)
def test_get_sorted_qs(data, expected):
    assert get_sorted_qs(data) == expected


def test_get_private_key():
    # from hex str
    pk = get_private_key(PRIVATE_KEY)
    assert dump_buf(pk.to_bytes(), dump=True) == PRIVATE_KEY

    pk_hex_str = remove_prefix_0x(pk.to_hex())
    pk2 = get_private_key(pk_hex_str)
    assert dump_buf(pk2.to_bytes(), dump=True) == PRIVATE_KEY

    pk3 = get_private_key(pk2)
    assert dump_buf(pk3.to_bytes(), dump=True) == PRIVATE_KEY

    # from eth_keys.datatypes.PrivateKey
    assert pk3 == get_private_key(pk3)

    # from bytes
    pk4 = get_private_key(hex_to_buf(PRIVATE_KEY))
    assert dump_buf(pk4.to_bytes(), dump=True) == PRIVATE_KEY

    # others raise ValueError
    with pytest.raises(ValueError):
        assert get_private_key(49811479637078589373593593025073956895140807052997722971398738500270349278662)


@pytest.mark.parametrize('dump', [True, False])
def test_recover_private_key(dump):
    keystore = json.dumps(KEYSTORE)
    private_key = recover_private_key(keystore, PASSWORD, dump)
    if dump:
        assert isinstance(private_key, str)
        assert private_key == PRIVATE_KEY
    else:
        assert isinstance(private_key, bytes)
        assert private_key == hex_to_buf(PRIVATE_KEY)


def test_private_key_to_address():
    address = private_key_to_address(PRIVATE_KEY)
    assert address == ADDRESS


def test_keccak256():
    message = 'hello prs'
    _hash = keccak256(message)
    _HASH = '647df39ad889e83cc0b9b65375672d1bfe282565c564d3d553a435bf80e67d92'
    assert _hash == _HASH


@pytest.mark.parametrize('dump', [True, False])
def test_create_key_pair(dump):
    pair = create_key_pair(dump)
    for item in {'privateKey', 'publicKey', 'address'}:
        assert item in pair
        v = pair[item]
        if item == 'address':
            assert isinstance(v, str)
            assert v[:2].lower() != '0x'
            continue

        if dump:
            assert isinstance(v, str)
            assert v[:2].lower() != '0x'
        else:
            assert isinstance(v, bytes)


@pytest.mark.parametrize(
    '_hash, expected',
    [
        (
            '565b63ac79b7d35a05322975340ae243e35ce084ae285c719fa6b203916f2845',
            'c47b72a8c7fa6407027deba86d3466f19c3546970214e8aa0b58094d4de4043a1c1c4893098d83b692528b678837d218230a8d0d5aaff4d4b42d50b255cd774e00'
        ),
        (
            'ab2fda04d97bba54f4d45e5b86ff62c5b720c8adbac277ee5920bf916b735f28',
            '9cb66fa967e970129569e8b164785edf183e76a4d3cdffecf1f918a1fa7835ce01ab4c2ee72e8aa1a5fae32c5920b0e01a97f104a490b4afca57c0fde664ab1300'
        ),
        (
            'e196055fda057d4abcab0a1b1b0ed54d5a33d23ad348e903bbaacf6c95d8404e',
            'dd9a6dc3352bd1c864a2c626acd075f8b7c48be30bdc9c9d4d316c7b02bf038f2fb1f6c9e1b7868c75e10abb76cf2eceab1eca1002897ccc6b59a33d0512af1701'
        )
    ]
)
def test_sign_hash(_hash, expected):
    private_key = '6e204c62726a19fe3f43c4ca9739b7ffa37e4a3226f824f3e24e00a5890addc6'
    sign = sign_hash(_hash, private_key)
    assert sign['signature'] == expected


def test_sign_text():
    key_pair = create_key_pair()
    private_key = key_pair['privateKey']
    address = key_pair['address']
    message = 'hello, world'
    sig = sign_text(message, private_key)
    _hash, signature = sig['hash'], sig['signature']
    recover_addr = sig_to_address(_hash, signature)
    assert address == recover_addr


def test_sign_block_data():
    key_pair = create_key_pair()
    private_key = key_pair['privateKey']
    address = key_pair['address']
    data = {'a': 111, 'b': 222}
    sig = sign_block_data(data, private_key)
    _hash, signature = sig['hash'], sig['signature']
    recover_addr = sig_to_address(_hash, signature)
    assert address == recover_addr


def test_sig_to_address_from_block():
    key_pair = create_key_pair()
    private_key = key_pair['privateKey']
    address = key_pair['address']
    data = {'a': 111, 'b': 222}
    sig = sign_block_data(data, private_key)
    signature = sig['signature']
    recover_addr = sig_to_address_from_block(data, signature)
    assert address == recover_addr



import io
import os
import re
from collections import OrderedDict
from setuptools import setup

here = os.path.abspath(os.path.dirname(__file__))


def read(*parts):
    with io.open(os.path.join(here, *parts), 'rt') as fp:
        return fp.read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(
        r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M
    )
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


readme = read('README.md')
version = find_version('prs_utility', '__init__.py')
github_url = 'https://github.com/Press-One/prs-utility-py'

setup(
    name="prs-utility",
    version=version,
    url=github_url,
    license='MIT',
    author="PRESS.one",
    author_email="dev@press.one",
    description="prs utility",
    long_description=readme,
    long_description_content_type="text/markdown",
    packages=['prs_utility'],
    platforms='any',
    python_requires='>=3.6',
    install_requires=[
        'eth-keyfile',
        'coincurve',
        'eth-utils',
        'eth-keys',
    ],
    extras_require={
        'dev': [
            'mypy',
            'pytest>=3',
            'coverage',
            'tox',
            'twine',
        ],
    },
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)
